/* SPDX-License-Identifier: GPL-2.0-or-later */
/*
 * Copyright (C) 2013 Red Hat, Inc.
 */

#ifndef NMTUI_HOSTNAME_H
#define NMTUI_HOSTNAME_H

NmtNewtForm *nmtui_hostname(gboolean is_top, int argc, char **argv);

#endif /* NMTUI_HOSTNAME_H */
