/* SPDX-License-Identifier: LGPL-2.1-or-later */

#ifndef __NM_META_SETTING_BASE_H__
#define __NM_META_SETTING_BASE_H__

#define _NM_META_SETTING_BASE_IMPL_LIBNM 0

#include "nm-meta-setting-base-impl.h"

#endif /* __NM_META_SETTING_BASE_H__ */
